# rcdl/scripts/migrate_old_format_to_db.py

"""
Use this script to migrate from pre-2.0 to 2+.
Script migrate_old_format_to_db version: v1.0

old_format:
- no db
cdl/
    creator1/
        date_title1_p0.mp4
        date_title2_p0.mp4
        date_title2_p1.mp4
        ...
    ...

for each creator:
    - [x] get all posts
    - [x] match each videos to its posts
    - [x] update db
    - [ ] check db with local videos
"""

import logging
import os
from pathlib import Path
import shutil

from rcdl.core.config import Config
from rcdl.core.models import Creator, Video, VideoStatus
from rcdl.core.file_io import load_json
from rcdl.core.api import URL
from rcdl.core.db import DB
import rcdl.core.downloader as dl
import rcdl.core.parser as parser

Config.ensure_dirs()
Config.ensure_files()

with DB() as db:
    db.init_table()

CDL_PATH = Path.home() / "Videos" / "cdl"
CREATORS_JSON = CDL_PATH / ".cache" / "creators.json"
TEMP_JSON = str(CDL_PATH / "temp.json")
LOG_PATH = CDL_PATH / "migrate_log.log"
MV_TXT = CDL_PATH / "mv.txt"
DEST_BASE = Path("/home/elitedesk/Videos/rcdl")

logging.basicConfig(
    filename=LOG_PATH,
    filemode="a",
    encoding="utf-8",
    level=logging.INFO,
    format="{asctime} - {levelname} - {message}",
    style="{",
    datefmt="%Y-%m-%d %H:%M:%S",
)
console = logging.StreamHandler()
console.setFormatter(logging.Formatter("{levelname}: {message}", style="{"))
logging.getLogger().addHandler(console)

open(MV_TXT, "w").close()


def add_to_mvtxt(filepath: str):
    with open(MV_TXT, "a") as f:
        f.write(filepath + "\n")


def update_db_info(video: Video):
    with DB() as db:
        db._upsert_video(video)


def move_files(mv_txt: Path, dest_base: Path):
    # Read all file paths
    i = 0
    with mv_txt.open("r", encoding="utf-8") as f:
        paths = [line.strip() for line in f if line.strip()]

    for src_path_str in paths:
        src = Path(src_path_str)
        if not src.exists():
            print(f"Skipping missing file: {src}")
            i += 1
            continue

        # Compute destination folder
        relative_parts = src.parts[len(Path("/home/elitedesk/Videos/cdl").parts) :]
        dest_dir = dest_base.joinpath(*relative_parts[:-1])
        dest_dir.mkdir(parents=True, exist_ok=True)

        # Destination path
        dest = dest_dir / src.name

        # Move the file
        shutil.move(str(src), str(dest))
        i += 1
        print(f"Moved ({i}/{len(paths)}): {src} -> {dest}")


if __name__ == "__main__":
    logging.info("--- MIGRATE pre-v2 to +2.0 Script START ---")

    # get all local creators
    creators_json = load_json(CREATORS_JSON)
    creators: list[Creator] = []
    for creator in creators_json:
        creators.append(
            Creator(
                creator_id=creator["creator_id"],
                service=creator["service"],
                domain=creator["domain"],
                status=None,
            )
        )

    for creator in creators:
        # get posts
        url = URL.get_creator_post_wo_param(creator)
        print(f"Request {url} up to {15} max page")
        pf = dl.PostsFetcher(url, TEMP_JSON, max_page=15)
        pf.request()

        posts = parser.filter_posts_with_videos_from_json(TEMP_JSON)
        print(f"Found {len(posts)} posts with videos for creator {creator.creator_id}")

        posts_videos = parser.convert_posts_to_videos(posts)
        print(f"Converted {len(posts)} to Video")
        for video in posts_videos:
            video.status = VideoStatus.DOWNLOADED

        # get local videos
        creator_path = os.path.join(CDL_PATH, creator.creator_id)
        print(f"Looking in {creator_path}")
        files = os.listdir(creator_path)

        local_videos: list[dict] = []
        for file in files:
            # ignore partial file
            if file.endswith(".part") or file.endswith(".aria2"):
                continue

            # remove ext
            name = file[:-4] if file.endswith(".mp4") else file

            # part number
            if "_p" not in name:
                print(f"Skipped file due to missing part number: {file}")

            try:
                base, part_str = name.rsplit("_p", 1)

                # extract date and title
                date, *title_parts = base.split("_")
                title = "_".join(title_parts)  # keeps underscores in title

                local_videos.append(
                    {
                        "date": date,  # e.g. "2025-12-25"
                        "title": title,
                        "part": part_str,  # e.g. "0"
                    }
                )
            except:  # noqa E277
                pass

        print(f"Found {len(local_videos)} local videos")

        # match local vid to videos list
        for lv in local_videos:
            for pv in posts_videos:
                if (
                    lv["date"] == pv.published
                    and lv["title"] == pv.title
                    and lv["part"] == str(pv.part)
                ):
                    print(f"Found a match for {pv.relative_path}")
                    update_db_info(pv)
                    add_to_mvtxt(os.path.join(creator_path, pv.relative_path))
                    break
            else:
                print(f"No match found {lv['date']}_{lv['title']}_{lv['part']}")

        move_files(MV_TXT, DEST_BASE)
        shutil.move(str(MV_TXT), str(MV_TXT) + f".{creator.creator_id}.txt")
        open(MV_TXT, "w").close()



# core/config.py

from pathlib import Path
import json
import logging
import os

from .file_io import write_json


class Config:
    # paths
    APP_NAME = "cdl"

    BASE_DIR = Path(os.environ.get("RCDL_BASE_DIR", Path.home() / "Videos/rcdl"))

    CACHE_DIR = BASE_DIR / ".cache"
    DB_PATH = CACHE_DIR / "cdl.db"
    LOG_FILE = CACHE_DIR / "cdl.log"
    FUSE_CSV_FILE = CACHE_DIR / "cdl_fuse.csv"
    SETTINGS_FILE = CACHE_DIR / "settings.json"
    CREATORS_FILE = CACHE_DIR / "creators.json"
    DISCOVER_DIR = CACHE_DIR / "discover"

    # defaults
    DEFAULT_SETTINGS = {
        "work_folder": str(BASE_DIR),
        "yt_dlp_args": "--external-downloader aria2c",
        "preset": "veryfast",
    }

    # default creators
    DEFAULT_CREATORS = [
        {"creator_id": "boixd", "service": "onlyfans", "domain": "coomer"}
    ]

    DEBUG = False
    DRY_RUN = False

    # api settings
    POST_PER_PAGE = 50
    DEFAULT_MAX_PAGE = 10
    MAX_FAIL_COUNT = 7

    @classmethod
    def ensure_dirs(cls):
        cls.CACHE_DIR.mkdir(parents=True, exist_ok=True)
        cls.DISCOVER_DIR.mkdir(exist_ok=True)

    @classmethod
    def ensure_files(cls):
        files = [
            cls.DB_PATH,
            cls.FUSE_CSV_FILE,
            cls.CREATORS_FILE,
        ]
        for file in files:
            if not file.exists():
                file.touch()
                logging.info("Created file %s", file)
                if file == cls.CREATORS_FILE:
                    write_json(file, cls.DEFAULT_CREATORS)

    @classmethod
    def load_settings(cls):
        if not cls.SETTINGS_FILE.exists():
            with open(cls.SETTINGS_FILE, "w") as f:
                json.dump(cls.DEFAULT_SETTINGS, f, indent=4)
            return cls.DEFAULT_SETTINGS

        with open(cls.SETTINGS_FILE, "r") as f:
            return json.load(f)

    @classmethod
    def creator_folder(cls, creator_id: str) -> Path:
        folder = cls.BASE_DIR / creator_id
        folder.mkdir(exist_ok=True)
        return folder

    @classmethod
    def cache_file(cls, filename: str, ext: str = ".json") -> Path:
        file_name = filename + ext
        file = cls.CACHE_DIR / file_name
        return file

    @classmethod
    def set_debug(cls, debug: bool):
        cls.DEBUG = debug

    @classmethod
    def set_dry_run(cls, dry_run: bool):
        cls.DRY_RUN = dry_run


def setup_logging(log_file: Path, level: int = 0):
    logger = logging.getLogger()
    logger.setLevel(level)
    logger.handlers.clear()  # avoid double handlers if called multiple times

    # file handler
    file_handler = logging.FileHandler(log_file, encoding="utf-8", mode="a")
    file_handler.setFormatter(
        logging.Formatter(
            "{asctime} - {levelname} - {message}",
            style="{",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    )
    logger.addHandler(file_handler)

    # console handler (stdout)
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(logging.Formatter("{levelname}: {message}", style="{"))
    logger.addHandler(console_handler)



# core/downloader.py

import logging
import os

import requests

import rcdl.core.parser as parser
from .api import URL
from .config import Config
from .models import Creator, Video, VideoStatus
from .db import DB
from .downloader_subprocess import ytdlp_subprocess
from .file_io import write_json, load_json
from rcdl.interface.progress import ProgressPrinter


class PostsFetcher:
    def __init__(
        self, url: str, json_path: str, max_page: int = Config.DEFAULT_MAX_PAGE
    ):
        self.url = url
        self.json_path = json_path

        self.page = 0
        self.max_page = max_page

        self.status = 200

    def _request_page(self, url: str) -> requests.Response:
        logging.info(f"RequestEngine url {url}")
        headers = URL.get_headers()
        return requests.get(url, headers=headers)

    def request(self, continue_on_error: bool = False, params: dict = {}):
        while self.status == 200 and self.page < self.max_page:
            o = self.page * Config.POST_PER_PAGE
            params["o"] = o
            url = URL.add_params(self.url, params)

            if Config.DRY_RUN:
                logging.debug(f"DRY-RUN posts fetcher {url} -> {self.json_path}")
                self.page += 1
                continue

            response = self._request_page(url)
            self.status = response.status_code

            if self.status != 200:
                if self.page == 0:
                    logging.error(f"Failed to get {url}")
                else:
                    logging.warning(
                        f"Status code {self.status}. This behavior is expected."
                    )
                if not continue_on_error:
                    break
            else:
                logging.info(f"Response Status Code: {self.status}")
                if self.page > 0:
                    json_data = list(load_json(self.json_path))
                else:
                    json_data = []

                if "posts" in response.json():
                    json_data.extend(response.json()["posts"])
                else:
                    json_data.extend(response.json())

                write_json(self.json_path, json_data, mode="w")

            self.page += 1


class VideoDownloader:
    def __init__(self, total_dl: int = -1):
        self.progress = ProgressPrinter(total_dl)

    def _build_url(self, domain: str, video: Video):
        return URL.get_url_from_file(domain, video.url)

    def _build_output_path(self, video: Video, discover: bool = False):
        if not discover:
            return os.path.join(
                Config.creator_folder(video.creator_id), video.relative_path
            )
        else:
            return os.path.join(Config.DISCOVER_DIR, video.relative_path)

    def _update_db_status(self, result: int, video: Video):
        with DB() as d:
            if result == 0:
                d.set_status(video, VideoStatus.DOWNLOADED)
            else:
                d.set_status(video, VideoStatus.FAILED, fail_count=video.fail_count + 1)

    def _exists(self, path: str) -> bool:
        if os.path.exists(path):
            return True
        return False

    def download(
        self, domain: str, video: Video, write_db: bool = True, discover: bool = False
    ) -> bool:
        url = self._build_url(domain, video)
        path = self._build_output_path(video, discover=discover)

        if Config.DRY_RUN:
            logging.debug(f"DRY-RUN dl {url} -> {path}")
            return True

        if self._exists(path):
            logging.warning("Video was already dl ! This should not happen")
            self.progress.update(ignore=True)  # update progress but not eta

        result = ytdlp_subprocess(url, path)
        if write_db:
            self._update_db_status(result, video)

        self.progress.update()
        self.progress.display()

        if result == 0:
            return True
        return False


def fetch_posts_by_tag(tag: str, max_page: int = Config.DEFAULT_MAX_PAGE) -> dict:
    url = URL.get_posts_page_url_wo_param()
    path = Config.cache_file(tag)
    pf = PostsFetcher(url, str(path), max_page=max_page)
    pf.request(continue_on_error=True, params={"tag": tag})

    return load_json(path)


def fetch_posts_by_creator(creator: Creator) -> dict:
    url = URL.get_creator_post_wo_param(creator)
    path = Config.cache_file(f"{creator.creator_id}_{creator.service}")
    pf = PostsFetcher(url, str(path))
    pf.request()

    return load_json(path)


def refresh_creators_videos(creators: list[Creator]):
    for creator in creators:
        logging.info(
            f"CREATOR {creator.creator_id} from {creator.service} on {creator.domain}"
        )

        posts = fetch_posts_by_creator(creator)
        logging.info(
            f"Found {len(posts)} posts from creator {creator.creator_id}({creator.service})"
        )
        posts_with_videos = parser.filter_posts_with_videos_from_json(
            str(Config.cache_file(f"{creator.creator_id}_{creator.service}"))
        )
        logging.info(
            f"Find {len(posts_with_videos)} posts with videos from creator {creator.creator_id}({creator.service})"
        )
        all_videos = parser.convert_posts_to_videos(posts_with_videos)
        logging.info(f"Converted {len(posts_with_videos)} to {len(all_videos)} videos")

        # put all videos in db
        with DB() as db:
            db.insert_videos(all_videos)


def download_videos_to_be_dl():
    """
    Download videos of all creators in creators.json
    """
    with DB() as db:
        videos = db.query_videos(pending=True)
        # VideoDownloader Engine
        vd = VideoDownloader(total_dl=len(videos))

        for video in videos:
            creator = Creator(
                creator_id=video.creator_id,
                service=video.service,
                domain=video.domain,
                status=None,
            )
            succesful = vd.download(creator.domain, video)
            if not succesful:
                logging.warning("Fail to dl vid")


def discover(tag: str, max_page: int):
    discover_creators(tag, max_page)
    dl_video_from_discover_creators()


def discover_creators(tag: str, max_page: int):
    # download posts with searched tags
    posts = fetch_posts_by_tag(tag, max_page)
    logging.info(f"Find {len(posts)} post")

    path = str(Config.cache_file(tag))
    posts_with_videos = parser.filter_posts_with_videos_from_json(path)
    logging.info(f"Find {len(posts_with_videos)} posts with videos")

    creators = parser.get_creators_from_posts(posts_with_videos)

    # save to csv
    file = os.path.join(Config.DISCOVER_DIR, "discover.csv")
    with open(file, "w") as f:
        for c in creators:
            line = f"{c.creator_id};{c.service};{c.domain};{'to_be_treated'}\n"
            f.write(line)


def dl_video_from_discover_creators():
    # load csv
    file = os.path.join(Config.DISCOVER_DIR, "discover.csv")
    with open(file, "r") as f:
        lines = f.readlines()

    creators = []
    for line in lines:
        line = line.replace("\n", "").strip().split(";")
        creators.append(
            Creator(creator_id=line[0], service=line[1], domain=line[2], status=line[3])
        )

    # get posts
    for creator in creators:
        response = requests.get(
            URL.get_creator_post_wo_param(creator), headers=URL.get_headers()
        )
        if response.status_code != 200:
            print(f"ERROR - Request {URL.get_creator_post_wo_param(creator)}")
        response_posts = response.json()
        posts = parser.filter_posts_with_videos_from_list(response_posts)
        print(f"{len(posts)} found")
        if len(posts) > 5:
            posts = posts[0:5]
            print("Limited posts to 5")

        for post in posts:
            urls = parser.extract_video_urls(post)
            url = URL.get_url_from_file(creator.domain, urls[0])
            filename = f"{post['user']}_{post['id']}.mp4"
            filepath = os.path.join(Config.DISCOVER_DIR, filename)
            ytdlp_subprocess(url, filepath)



# core/models.py

from dataclasses import dataclass
from typing import Optional
from enum import Enum


class VideoStatus(Enum):
    NOT_DOWNLOADED = "not_downloaded"
    DOWNLOADED = "downloaded"
    FAILED = "failed"
    SKIPPED = "skipped"
    IGNORED = "ignored"


class DiscoverStatus(Enum):
    TO_BE_TREATED = "to_be_treated"
    DOWNLOADED = "downloaded"
    BLACKLISTED = "blacklisted"
    WHITELSITED = "whitelisted"
    DOWNLOAD_MORE = "download_more"


@dataclass
class Creator:
    creator_id: str
    service: str
    domain: str
    status: Optional[str]


@dataclass
class Video:
    # important fields
    post_id: str
    creator_id: str
    service: str
    domain: str
    relative_path: str
    url: str
    part: int = 0

    # metadata
    published: Optional[str] = None
    title: Optional[str] = None
    substring: Optional[str] = None
    downloaded_at: Optional[str] = None
    file_size: Optional[float] = None

    # status in cdl
    status: Optional[VideoStatus] = None
    fail_count: int = 0



# core/parser.py

import logging
from pathvalidate import sanitize_filename

from .models import Video, VideoStatus, Creator
from .db import DB
from .file_io import load_json
from .config import Config


COOMER_PAYSITES = ["onlyfans", "fansly", "candfans"]


def get_domain(arg) -> str:
    def _service(service: str) -> str:
        if service in COOMER_PAYSITES:
            return "coomer"
        return "kemono"

    def _get_domain_from_post(post: dict) -> str:
        return _service(post["service"])

    if isinstance(arg, dict):
        return _get_domain_from_post(arg)

    return _service(arg)


def get_title(post: dict) -> str:
    title = post["title"]
    if title == "":
        if "content" in post:
            title = post["content"]
        elif "substring" in post:
            title = post["substring"]
    return sanitize_filename(title)


def get_date(post: dict) -> str:
    if "published" in post:
        date = post["published"][0:10]
    elif "added" in post:
        date = post["added"][0:10]
    else:
        logging.error(f"Could not extract date from {post['id']}")
        date = "NA"
    return date


def get_part(post: dict, url: str) -> int:
    urls = extract_video_urls(post)
    part = 0
    if len(urls) == 1:
        return 0

    for u in urls:
        if u == url:
            return part
        part += 1

    logging.error(
        f"Could not extract part number for post id {post['id']} with url {url}"
    )
    return -1


def get_filename(post: dict, url: str) -> str:
    title = get_title(post)
    date = get_date(post)
    part = get_part(post, url)
    file_title = f"{date}_{title}".replace("'", " ").replace('"', "")
    filename = f"{file_title}_p{part}.mp4"
    return filename


def convert_post_to_video(post: dict, url: str, discover=False) -> Video:
    part = get_part(post, url)
    title = get_title(post)
    date = get_date(post)
    filename = get_filename(post, url)

    if discover:
        filename = f"{post['user']}_{post['id']}.mp4"

    return Video(
        post_id=post["id"],
        creator_id=post["user"],
        service=post["service"],
        relative_path=filename,
        url=url,
        domain=get_domain(post),
        part=part,
        published=date,
        title=title,
        status=VideoStatus.NOT_DOWNLOADED,
        fail_count=0,
    )


def convert_posts_to_videos(posts: list[dict], discover: bool = False) -> list[Video]:
    videos = []
    for post in posts:
        urls = extract_video_urls(post)
        if not discover:
            for url in urls:
                videos.append(convert_post_to_video(post, url))
        else:
            if len(urls) == 0:
                continue
            videos.append(convert_post_to_video(post, urls[0], discover=True))
    return videos


def extract_video_urls(post: dict) -> list:
    video_extensions = (".mp4", ".webm", ".mov", ".avi", ".mkv", ".flv", ".wmv", ".m4v")
    urls = set()

    # Check main file
    if post["file"]:
        if post["file"]["path"]:
            path = post["file"]["path"]
            if path.endswith(video_extensions):
                urls.add(f"{path}")

    if post["attachments"]:
        attachments = post["attachments"]
        for attachment in attachments:
            if attachment["path"]:
                if attachment["path"].endswith(video_extensions):
                    urls.add(f"{attachment['path']}")

    return list(urls)


def filter_posts_with_videos_from_list(data: list[dict]) -> list[dict]:
    """Return posts with video url from a json with a list of posts"""

    posts_with_videos = []
    for post in data:
        if len(extract_video_urls(post)) > 0:
            posts_with_videos.append(post)
    return posts_with_videos


def filter_posts_with_videos_from_json(path: str) -> list:
    """Return posts with video url from a json with a list of posts"""
    posts = load_json(path)

    posts_with_videos = []
    for post in posts:
        if len(extract_video_urls(post)) > 0:
            posts_with_videos.append(post)
    return posts_with_videos


def filter_undownloaded_videos(videos: list[Video]) -> list[Video]:
    with DB() as db:
        downloaded_videos = db.query_videos(status=VideoStatus.DOWNLOADED)

    downloaded_keys = set()
    for v in downloaded_videos:
        downloaded_keys.add((v.service, v.url))

    videos_to_dl = []
    for v in videos:
        key = (v.service, v.url)
        if key not in downloaded_keys:
            videos_to_dl.append(v)
    return videos_to_dl


def _convert_creator_dict_to_creator(creator: dict):
    return Creator(
        creator_id=creator["creator_id"],
        service=creator["service"],
        domain=creator["domain"],
        status=None,
    )


def get_creators() -> list[Creator]:
    data = load_json(Config.CREATORS_FILE)
    creators = []
    for creator in data:
        creators.append(_convert_creator_dict_to_creator(creator))
    return creators


def get_creators_from_posts(posts: list[dict]) -> list[Creator]:
    creators = list()
    seen = set()

    for post in posts:
        key = (post["user"], post["service"], "coomer")
        if key in seen:
            continue

        seen.add(key)
        creators.append(
            Creator(
                creator_id=post["user"],
                service=post["service"],
                domain="coomer",
                status="to_be_treated",
            )
        )
    return creators



# core/processor.py

import logging
import os
import time

from moviepy import VideoFileClip, concatenate_videoclips

from rcdl.utils import echo, echoc, Progress
from .config import Config


def rm_file(f):
    try:
        os.remove(f)
        logging.info(f"removed file {f}")
    except Exception as e:
        logging.error(f"Failed to delete {f} due to: {e}")


def rename_moved_file(f):
    dir_name = os.path.dirname(f)
    file_name = os.path.basename(f)

    new_file_name = f".mv_{file_name}"
    new_path = os.path.join(dir_name, new_file_name)
    try:
        os.rename(f, new_path)
        logging.info(f"renamed file {f}->{new_path}")
    except Exception as e:
        logging.error(f"Failed to rename {f}->{new_path} due to {e}")


def fuse_csv(
    filename: str,
    threads: int,
    preset: str,
    number_of_parts: int,
    original_total_byte_size: int,
    result_byte_size: int,
    encoding_time: float,
):
    """Write info to fuse.csv for data gathering"""

    headers = [
        "filename",
        "threads",
        "preset",
        "number_of_parts",
        "original_total_byte_size",
        "result_byte_size",
        "encoding_time",
    ]
    headers_str = ""
    for h in headers:
        headers_str += h + ";"

    path = Config.FUSE_CSV_FILE
    if not os.path.exists(path):
        with open(path, "w") as f:
            f.write(headers_str + "\n")

    line = f"{filename};{threads};{preset};{number_of_parts};{original_total_byte_size};{result_byte_size};{encoding_time};\n"
    with open(path, "a") as f:
        f.write(line)


def find_all_parts_video(folder_path: str) -> list:
    """Find all video that are _p1, _p2 etc.., extract the title from [0 to _pX.mP4(not included)] and return a set"""
    groups = {}

    videos_files = os.listdir(folder_path)
    for video in videos_files:
        if video.startswith(".mv_"):
            continue
        if not video.endswith(".mp4"):
            continue

        name = video[:-4]  # remove .mp4
        if "_p" not in name:
            continue

        title, part = name.rsplit("_p", 1)
        if part.isdigit():
            part_num = int(part)
            groups.setdefault(title, []).append((part_num, video))

    result = []
    for title, files in groups.items():
        if len(files) <= 1:
            continue

        sorted_files = sorted(files, key=lambda x: x[0])

        filenames = [fname for _, fname in sorted_files]

        result.append(filenames)

    return result


def verify_parts_videos(groups: list):
    """Take a list of lists of parts videos, verify none are missing"""
    valid_groups = []

    for files in groups:
        part_numbers = []
        for f in files:
            name = f[:-4]  # remove .mp4

            if "_p" not in name:
                error_msg = (
                    f"Files {files} not valid due to file {f} missing '_p' in filename"
                )
                echoc(error_msg, "red")
                logging.error(error_msg)
                break  # invalid format

            _, part = name.rsplit("_p", 1)
            if not part.isdigit():
                error_msg = f"Files {files} not valid due to file {f} having a invalid part number (not a number)"
                echoc(error_msg, "red")
                logging.error(error_msg)
                break

            part_numbers.append(int(part))

        if not part_numbers:
            error_msg = f"Files {files} not valid due to empty part numbers"
            echoc(error_msg, "red")
            logging.error(error_msg)
            continue  # skip invalid/empty groups

        part_numbers.sort()
        min_p, max_p = 0, part_numbers[-1]
        expected = list(range(min_p, max_p + 1))
        if part_numbers == expected and min_p == 0:
            valid_groups.append(files)
        else:
            missing_parts = set(expected) - set(part_numbers)
            error_msg = f"Files {files} not valid due to a mismatch on the number of part: expected {len(expected)}, got {len(part_numbers)}\nMissing part: _p{missing_parts}"
            echoc(error_msg, "red")
            logging.error(error_msg)

    return valid_groups


def fuse_videos(groups: list, creator_folder_path: str):
    index = 0
    progress = Progress(len(groups))

    for videos_files in groups:
        # echo(f"[cdl] {progress.get_str()}")
        # get full absolute path
        for i in range(len(videos_files)):
            videos_files[i] = creator_folder_path + videos_files[i]

        # extract title for output
        filepath = str(videos_files[0])
        filepath = filepath[:-4]  # remove .mp4
        filepath_part = filepath.rsplit("_p", 1)
        filepath = filepath_part[0]
        filepath += ".mp4"

        # output filepath
        output_path = filepath
        if Config.DEBUG:
            output_path = f"dev/output_{index}.mp4"

        clips = []
        final = None

        try:
            # add clips
            for f in videos_files:
                clip = VideoFileClip(f)
                clips.append(clip)

            # concatenate them indepened of codec, sizes, resolution
            final = concatenate_videoclips(
                clips, method="compose"
            )  # "compose" handles different sizes/codecs

            # select number of threads
            threads = os.cpu_count()
            if threads is None:
                threads = 2

            # load preset from settings
            valid_preset = [
                "ultrafast",
                "superfast",
                "veryfast",
                "faster",
                "fast",
                "medium",
                "slow",
                "slower",
                "veryslow",
                "placebo",
            ]
            preset = Config.load_settings()["preset"]
            if preset not in valid_preset:
                error_msg = f"Preset {preset} is not valid. Valid preset are: {valid_preset}. Please check settings.json"
                logging.error(error_msg)
                echoc(error_msg, "red")
                return

            # logging info var
            parts_number = len(videos_files)
            total_size = 0  # in bytes
            for f in videos_files:
                total_size += os.path.getsize(f)
            start_time = time.time()

            msg = f"Concatenating {filepath} with {threads} threads"
            echo(msg)
            logging.info(msg)

            final.write_videofile(output_path, preset=preset, threads=threads)

            # log info data
            result_size = os.path.getsize(output_path)
            encoding_time = time.time() - start_time
            fuse_csv(
                filepath,
                threads,
                preset,
                parts_number,
                total_size,
                result_size,
                encoding_time,
            )

            msg = f"Result for {filepath}: threads: {threads}, preset: {preset}, nb_of_parts: {parts_number}, original_byte_size: {total_size}, result_size: {result_size}, time: {encoding_time}s"
            if Config.DEBUG:
                msg += f" --DEBUG: output_path != filepath: {output_path}"
            logging.info(msg)
            echo(msg)

        except Exception as e:
            error_msg = f"Failed to concatenate {filepath} due to e: {e}"
            echoc(error_msg, "red")
            logging.error(error_msg)
        else:
            # rm part file
            for f in videos_files:
                # rm_file(f)
                rename_moved_file(f)
        finally:
            # free memory
            for clip in clips:
                try:
                    clip.close()
                except Exception as e:
                    logging.warning(f"Failed to close clip: {e}")
            try:
                if final is not None:
                    final.close()
            except Exception as e:
                logging.warning(f"Failed to close VideoClip: {e}")

        index += 1
        progress.update()


def fuse_all_videos(creators: list):
    for creator in creators:
        creator_id = creator["creator_id"]
        creator_folder_path = str(Config.creator_folder(creator_id))

        echo(f"Finding part videos for {creator_id} in {creator_folder_path}")

        # find all part video
        groups = find_all_parts_video(creator_folder_path)
        msg = f"Find {len(groups)} video to fuse for creator {creator_id}"
        echo(msg)
        logging.info(msg)

        # validation of groups of part videos
        valid_groups = verify_parts_videos(groups)
        msg = f"Find {len(valid_groups)} valid groups to fuse; {len(groups) - len(valid_groups)} unvalid groups removed."
        echo(msg)
        logging.info(msg)

        # fuse vid
        if not Config.DRY_RUN:
            fuse_videos(valid_groups, creator_folder_path)
        else:
            echoc("DRY-RUN -> FUSE SKIPPED", "blue")



# downloader_subprocess.py

import subprocess
import logging
from pathlib import Path


def ytdlp_subprocess(
    url: str,
    filepath: Path | str,
):
    cmd = [
        "yt-dlp",
        "-q",
        "--progress",
        url,
        "-o",
        filepath,
        "--external-downloader",
        "aria2c",
    ]

    logging.info(f"CMD: {' '.join(cmd)}")

    result = subprocess.run(cmd, capture_output=True, text=True)
    if result.returncode != 0:
        logging.error(f"yt-dlp failed to dl vid: {result.stderr}")

    return result.returncode



# interface/progress.py

import click

from rcdl.utils import Progress


class ProgressPrinter:
    def __init__(self, total: int, progress=Progress):
        self.progress = progress(total)

    def update(self, ignore=False):
        self.progress.update(ignore=ignore)

    def _format_time(self, seconds: float) -> str:
        total_seconds = int(seconds)
        hours = total_seconds // 3600
        minutes = (total_seconds % 3600) // 60
        secs = total_seconds % 60
        return f"{hours:02d}:{minutes:02d}:{secs:02d}"

    def display(self):
        msg = f"Progress: {self.progress.current}/{self.progress.total} ({round(self.progress.pourcent, 1)}%)"
        msg += f" ETA: {self._format_time(self.progress.eta)}"
        click.echo(msg)



# core/file_io.py

import json


def write_json(path, data, mode="w"):
    with open(path, mode) as f:
        json.dump(data, f, indent=4)


def load_json(path) -> dict:
    with open(path, "r") as f:
        data = json.load(f)
    return data



# utils.py

import time

import click


class Progress:
    def __init__(self, total: int):
        self.total = total
        self.current = 0

        self.last_time = time.time()
        self.times = []

    def update(self, ignore=False):
        now = time.time()
        if not ignore:
            self.times.append(now - self.last_time)
        self.last_time = now
        self.current += 1

    @property
    def eta(self) -> float:
        if len(self.times) == 0:
            return 0.0
        avrg_time = sum(self.times) / len(self.times)
        return (self.total - self.current) * avrg_time

    @property
    def pourcent(self) -> float:
        return self.current / self.total * 100 if self.total else 0.0


def echo(text: str):
    click.echo(text)


def echoc(text: str, color: str):
    click.echo(click.style(text, fg=color))



# interface/cli.py

import logging
import os.path

import click

from rcdl.utils import echo, echoc
from rcdl.core import downloader as dl
from rcdl.core.config import Config
from rcdl.core.processor import fuse_all_videos
from rcdl.core.parser import get_creators
from rcdl.core.db import DB


from rcdl import __version__


@click.command(help="Refresh video to be downloaded")
def refresh():
    """
    Find all videos of creators
    """
    creators = get_creators()
    logging.info("[cdl] refresh")
    dl.refresh_creators_videos(creators)

    with DB() as db:
        info = db.get_db_videos_info()

    for key in ["not_downloaded", "downloaded", "failed", "skipped", "ignored"]:
        logging.info(f"{key}: {info[key]}")


@click.command(help="Download all videos from all creator")
def dlsf():
    """
    Download settings file
    Name was kept from pototype despite functionnality changing.
    """

    logging.info("[cdl] dlsf")
    dl.download_videos_to_be_dl()


@click.command(help="Show log info")
def log():
    """Show the log file in the terminal"""
    if not os.path.exists(Config.LOG_FILE):
        echoc(f"Logging file {Config.LOG_FILE} does not exist", "red")
        return

    try:
        with open(Config.LOG_FILE, "r") as f:
            logs = f.read()

        os.environ["LESS"] = os.environ.get("LESS", "") + " +G"
        click.echo_via_pager(logs)

    except Exception as e:
        echoc(f"Failed to display log content due to error: {e}", "red")


@click.command(help="Fuse multiple part video into one")
def fuse():
    """Fuse all videos of creators that are in multiple parts"""
    echoc("[cdl] fuse", "white")
    logging.info("[cdl] fuse")

    creators = get_creators()
    if Config.DEBUG and len(creators) > 1:
        creators = creators[0:1]

    fuse_all_videos(creators)


@click.command(help="Discover videos/creators with tags")
@click.option("--tag", required=True, type=str, help="Tag to search for")
@click.option(
    "--max-page", default=10, type=int, help="Maximum number of pages to fetch"
)
def discover(tag, max_page):
    """Discover new creators/videos"""
    msg = f"[cdl] discover with tag={tag} max_page={max_page}"
    echo(msg)
    logging.info(msg)
    dl.discover(tag, max_page)


# --- CLI GROUP ---
@click.group()
@click.option("--debug", is_flag=True)
@click.option("--dry-run", is_flag=True)
@click.version_option(version=__version__, prog_name=Config.APP_NAME)
def cli(debug, dry_run):
    Config.set_debug(debug)
    Config.set_dry_run(dry_run)


cli.add_command(dlsf)
cli.add_command(log)
cli.add_command(fuse)
cli.add_command(discover)
cli.add_command(refresh)
